/*
 * microcode_ctl - Manipulate /dev/cpu/microcode under Linux
 *
 * Copyright 2000 (c) Simon Trimmer, Tigran Aivazian.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

static char *progname;
static int print_normal_messages=1;
static int print_error_messages=1;

#define BUFFER_SIZE	4096
#define MICROCODE_SIZE (128*1024) /* initial size, expanded as needed */
#define FILENAME_MAXLEN 2048

#define MICROCODE_DEVICE_DEFAULT	"/dev/cpu/microcode"
#define MICROCODE_FILE_DEFAULT		"/usr/share/misc/microcode.dat"


static void usage(FILE *file)
{
	fputs(
"\nThis program is for updating the microcode on Intel processors\n"
"belonging to the IA32 family - PentiumPro upwards (x86_64 included).\n"
"It depends on the Linux kernel microcode driver.\n\n"
"Usage: microcode_ctl [-h] [-u] [-q] [-Q] [-f microcode]\n\n"
"  -h	this usage message\n"
"  -q	run silently when successful\n"
"  -Q	run silently even on failure\n"
"  -u	upload microcode (default filename:\""MICROCODE_FILE_DEFAULT"\"\n"
"  -f	upload microcode from named Intel formatted file\n\n",
        file);
}

/* 
 * The update has two stages; 
 * a) read in the Intel microcode file and convert it into a format suitable
 *    for the processor.
 * b) send the microcode to the driver which applies the update
 */
static int do_update(char *device, char *filename)
{
	FILE *fp;
	char line_buffer[BUFFER_SIZE];
        size_t microcode_size = MICROCODE_SIZE/sizeof(unsigned);
        unsigned *microcode = malloc(microcode_size * sizeof(unsigned));
        size_t pos = 0;
	int outfd;
	int wrote, length;

       if(!microcode){
               if(print_error_messages)
                       fprintf(stderr, "%s: not enough memory\n", progname);
               return ENOMEM;
       }

	if( (fp=fopen(filename, "r")) == NULL){
		if(print_error_messages)
                       fprintf(stderr,
                               "%s: cannot open source file '%s' errno=%d (%s)\n",
				progname, filename, errno, strerror(errno));
		return errno;
	}

	while(fgets(line_buffer, BUFFER_SIZE, fp) != NULL) {
	         /*
		  * Expand microcode buffer if needed
                  */
               if(microcode_size < pos + 4) {
                       microcode_size *= 2;
                       microcode = realloc(microcode,
                                          microcode_size * sizeof(unsigned));
                       if (!microcode) {
                               if(print_error_messages)
                                       fprintf(stderr, "%s: not enough memory\n",
                                               progname);
                               fclose(fp);
                               return ENOMEM;
                       }
               }
		 /*
		  * Data lines will are of the form "%x, %x, %x, %x", therefore
		  * lines start with a 0
		  */
                if(*line_buffer == '0'){
                        int scanned;
                        scanned = sscanf(line_buffer, "%x, %x, %x, %x",
                                         microcode + pos,
                                         microcode + pos + 1,
                                         microcode + pos + 2,
                                         microcode + pos + 3);
                        if(scanned != 4) {
                                fprintf(stderr, "%s: %s: invalid file format\n",
                                        progname, filename);
                                fclose(fp);
                                return EINVAL;
                        }
                        pos += 4;
		}
	}

	fclose(fp);
        length = pos * sizeof(unsigned);

	if(print_normal_messages)
		fprintf(stderr, "%s: writing microcode (length: %d)\n",  progname, length);
	
	if((outfd = open(device, O_WRONLY)) == -1){
		if(print_error_messages)
			fprintf(stderr, "%s: cannot open %s for writing errno=%d (%s)\n",
				progname, device, errno, strerror(errno));
		return errno;
	}

	if( (wrote = write(outfd, microcode, length)) < 0){
		if(print_error_messages)
			fprintf(stderr, "%s: error writing to '%s' errno=%d (%s)\n"
					"%s: there may be messages from the driver in your system log.\n",
				progname, device, errno, strerror(errno), progname);
		close(outfd);
		return errno;
	}

	if((wrote == length) && print_normal_messages)
			fprintf(stderr, "%s: microcode successfuly written to %s\n",
			       progname, device);

	close(outfd);

	return errno;
}

int main(int argc, char *argv[])
{
	int c;
        static char device[FILENAME_MAXLEN+1] = MICROCODE_DEVICE_DEFAULT;
        static char filename[FILENAME_MAXLEN+1] = MICROCODE_FILE_DEFAULT;
	int upload=0;

	progname = argv[0];

	while (EOF != (c = getopt(argc, argv, "-hqQiud:f:"))) {
		switch(c) {
			case 'h':
				usage(stdout);
				exit(0);

			case 'q':
				print_normal_messages=0;
				break;

			case 'Q':
				print_error_messages=0;
				print_normal_messages=0;
				break;

			case 'd':
				strncpy(device, optarg, FILENAME_MAXLEN);
				break;

			case 'u': /* do a microcode upload */
				upload++;
				break;

			case 'f': /* set microcode file to optarg and upload */
				upload++;
                                strncpy(filename, optarg, FILENAME_MAXLEN);
				break;
		        case 'i':
	  		        fputs("microcode_ctl: flag '-i' is now obsolete",
     				      stderr);
				break;
			case '?':
                        case 1:
                                usage(stdout);
                                exit(2);

		}
	}

	if (upload) {
                if(do_update(device, filename))
                        exit(1);
	} else
		usage(stderr);

	return 0;
}
